/** 
 *   Site testing for SKA candidate.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <stdarg.h>
#include <unistd.h>
#include <malloc.h>
#include <math.h>
#include <gpib/ib.h>
#include "analyzer.h"
#include "my_stdio.h"
#include "toska.h"
#include "observ_table.h"

#define MY_IBWRT_OLD(dev, fmt, arg...)\
        if ( my_ibwrt(dev, fmt, ## arg) & ERR ){ return -1; } 

#define MY_IBWRT(dev, fmt, arg...)\
        if ( my_ibwrite_generic(ibwrt, dev, fmt, ## arg) & ERR ){return -1;} 

#define MY_IBWRTA(dev, fmt, arg...)\
        if ( my_ibwrite_generic(ibwrta, dev, fmt, ## arg) & ERR ){return -1;} 

#define  PUT_TITLE(dev)\
  	MY_IBWRT(dev,"PU;PA 200,205;TEXT|-- SKA Survey (IAR) --|;");


static int my_ibwrite_generic(
	int (*write_func)(int ud, const void *data, long num_bytes),
	int dev, const char *fmt, ...)
{
	va_list ap;
	int l = -1;
	int n, size = 80;
	char *buff = NULL;

	if ((buff = malloc (size)) == NULL) return l;
	while (1) {
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf (buff, size, fmt, ap);
		va_end(ap);
		if (n > -1 && n < size) {
			break;
		}
		else {
			/* Else try again with more space. */
			if (n > -1)    /* glibc 2.1 */
				size = n+1; /* precisely what is needed */
			else           /* glibc 2.0 */
				size *= 2;  /* twice the old size */
			if ((buff = realloc (buff, size)) == NULL)
				return l;
		}
	}

	if (buff != NULL) {
		PVERB(PINFO,"# Send: %s\n", buff);
#ifdef DEMO_MODE
		l = write(dev, buff, n);
		write(dev, "\n", 1);

#else
		l = write_func(dev, buff, n);
#endif

#ifdef VERBOSE_MODE

		if (l & ERR && !(l & TIMO)) {
			SYSLOG(LOG_ERR,"Error sending: %s [ibstat: 0x%x]\n", 
			      buff, l);
		}
#endif
		free(buff);
	}

	return l;
}


int open_analyzer(int pad)
{
	int dev;
	short listen;

#ifdef DEMO_MODE
	return 1;
#endif
	/* Automatic serial polling is !enabled. */
	ibconfig(AN_GPIB, IbcAUTOPOLL, 0);

	dev = ibdev(AN_GPIB, pad, 0, AN_TIME_OUT, 1, 0);
	if (dev < 0) {
	  return dev;
	}

	/* check if listener is present */
	ibln(AN_GPIB, pad, NO_SAD, &listen );
	if(ibsta & ERR){
		if(iberr == EBUS){
			SYSLOG(LOG_ERR,
			       "Analyzer not found in primary address %d",
			       pad);
			return -1;
		}
		SYSLOG(LOG_ERR,"Analyzer: init error [0x%x]", iberr);
		return -1;
	}
	if( ! listen ){
		SYSLOG(LOG_ERR,
		       "Analyzer not found in primary address %d",
		       pad);
		return -1;
	}
	/* IP: Instrument Preset */
	MY_IBWRT(dev, "IP;");
	PUT_TITLE(dev);
	/* SNGLS: Single Sweep */
	MY_IBWRT( dev, "SNGLS;TS;");

	return dev;
}

int analyzer_reset( int dev )
{
	PVERB(PERR,"analyzer_reset\n");
	/* IP: Instrument Preset */
	MY_IBWRT(dev, "IP;");
	PUT_TITLE(dev);
	/* SNGLS: Single Sweep */
	MY_IBWRT( dev, "SNGLS;TS;");

	return 0;
}


/* FIXME: cambiar los argumentos para que solo reciba lo que va a
   hacer, y entregue cada tramode AN_MAX_POINTS, para que se junten en
   otro lado. */
static int get_track( int dev, double *data )
{
	int i;
	char buff[(AN_MAX_POINTS << 3) + 2];
	char *np = buff;
        char *endp;

	MY_IBWRT(dev,"TRA?;");
	
#ifdef DEMO_MODE
	for(i=0;i<AN_MAX_POINTS-1;i++){
		sprintf(buff+i*8,"%7.2f,", rand() / (RAND_MAX+1.0) * -100.0);
	}
	sprintf(buff+i*8,"%7.2f%c", rand() / (RAND_MAX+1.0) * -100.0,0);
#else
	ibrd(dev, buff, (AN_MAX_POINTS << 3) + 2);
#endif
	/* bytes_read = ibcnt; */
	PVERB(PINFO,"read: %d bytes\n", ibcnt);
	
	for(i=0;i<AN_MAX_POINTS;i++){
		*(data+i) =  strtod(np, &endp);
		np = endp + 1;
		//printf("%7.2f\n", *(s->data+i));
        }
	
	return 0;
}


static int wait_end(int dev)
{
	int i;
	int sweep_on;
	int res;
	char status;
	//time_t now;


	sweep_on = 1;
	i = 0;
	do{
#ifdef DEMO_MODE
		ibsta = 0;
		status = LACS;
		res = 0;
#else
		res = ibrsp(dev, &status);
		PVERB(PINFO,
                  "# ibrsp->status(%d): 0x%x wait for 0x%x, ibsta: 0x%x\n", 
			i, status, LACS, res);
#endif
		if(res & ERR){
			SYSLOG(LOG_ERR,"ibsta & ERR\n");
			return -1;			
		}
		if( status & LACS){
			sweep_on = 0;
		}else{
			sleep(1);
		}
		if( i > 5) {
		  SYSLOG(LOG_ERR,"Sweep time expired.\n");
		  return -1;
		}
		i++;
	}while(sweep_on);
#ifndef DEMO_MODE
	sleep(1);
#endif

	return 0;
}


static int setup_observ(int dev, int mode, double start_freq,double stop_freq,
			int RBW, int sweep, int current_rep, int repetitions,
			int azimuth, int polarization)
{
	MY_IBWRT(dev,"CLRW TRA;");
	MY_IBWRT(dev,
		 "PU;PA 10,180;TEXT|Mode: %d, Azt: %3d%c, Rep: %d/%d    |;", 
		 mode, azimuth, polarization, current_rep, repetitions);
	MY_IBWRT(dev,"FA %8.6fGHZ;", start_freq);
	MY_IBWRT(dev,"FB %8.6fGHZ;", stop_freq);
	MY_IBWRT(dev,"RB %dKHZ;", RBW);
	MY_IBWRT(dev,"ST %dSC;", sweep);
	
	MY_IBWRT(dev, "RQS %d;", 
		 RQS_EOS|RQS_UKP|RQS_CC|RQS_HB);  /* End Of Sweep */
	MY_IBWRT(dev, "TDF P;");

	return 0;
}


int analyzer_get_observ( int dev, struct obs_setup *s)
{
	int i;
	int res;
	time_t now;
	struct modes *obs = s->obs;
	int sweep;
	double start_freq;
	double stop_freq;
	double freq_inc;
	int split = 1;

	sweep = (int) ceil((double) obs->sweep / (double) s->partition);
	start_freq = obs->start_freq;
	freq_inc = obs->stop_freq - obs->start_freq;
	if (sweep > SWEEP_MAX){
		split = (int) ceil((double) sweep / (double) SWEEP_MAX);
		sweep = (int) ceil((double) sweep / (double) split);
		freq_inc = (obs->stop_freq - obs->start_freq)/(double) split;
	}
	obs->sweep_obs = sweep * split;
	s->naxis1 = split * AN_MAX_POINTS;
	s->data = (double *) malloc(s->naxis1 * sizeof(double));
	if (s->data == NULL){
		PVERB(PERR,"Can not allocate memory for track\n");
		return -1;
	}
	for(i=0; i<split; i++) { /*.....*/
		stop_freq = start_freq + freq_inc;
		res = setup_observ(dev, obs->mode, start_freq, stop_freq,
				   obs->RBW,  sweep, s->current_rep, 
				   obs->repetitions, s->azimuth, 
				   s->polarization);
		
		if ( res != 0 ) return -1;
		
		time(&s->start);
		res = my_ibwrite_generic(ibwrt, dev, "TS;");
		time(&now);
		
		if ( res & ERR && !(res & TIMO) ) {
			return -1;
		}
#ifndef DEMO_MODE
		sleep_time = s->start - now + obs->sweep;
		if ( sleep_time > 0 ) {
			PVERB(PINFO,"sleeping %d sec.\n", sleep_time);
			sleep(sleep_time);
		}
#endif
		if( wait_end(dev) != 0 ) {
			return -1;
		}
		
		if( get_track(dev, s->data+(i*AN_MAX_POINTS)) != 0 ) {
			return -1;
		}
		
		/* ... */
		
		start_freq = stop_freq;
		
	}
	
	return 0;
}




#if 0
int analyzer_get_observ_orig( int dev, struct obs_setup *s)
{
	int i;
	int sweep_on;
	int sleep_time;
	int res;
	char status;
	char buff[(AN_MAX_POINTS << 3) + 2];
	char *np = buff;
        char *endp;
	time_t now;
	struct modes *obs = s->obs;

	MY_IBWRT(dev,"CLRW TRA;");
	MY_IBWRT(dev,
		 "PU;PA 10,180;TEXT|Mode: %d, Azt: %3d%c, Rep: %d/%d    |;", 
		 obs->mode, s->azimuth, s->polarization,
		 s->current_rep, obs->repetitions);
	MY_IBWRT(dev,"FA %8.6fGHZ;", obs->start_freq);
	MY_IBWRT(dev,"FB %8.6fGHZ;", obs->stop_freq);
	MY_IBWRT(dev,"RB %dKHZ;", obs->RBW);
	MY_IBWRT(dev,"ST %dSC;", obs->sweep);
	
	MY_IBWRT(dev, "RQS %d;", 
		 RQS_EOS|RQS_UKP|RQS_CC|RQS_HB);  /* End Of Sweep */
	MY_IBWRT(dev, "TDF P;");
	time(&s->start);
	res = my_ibwrite_generic(ibwrt, dev, "TS;");
	time(&now);

	if ( res & ERR && !(res & TIMO) ) {
		return -1;
	}
#ifndef DEMO_MODE
	sleep_time = s->start - now + obs->sweep;
	if ( sleep_time > 0 ) {
	  PVERB(PINFO,"sleeping %d sec.\n", sleep_time);
	  sleep(sleep_time);
	}
#endif
	sweep_on = 1;
	i = 0;
	do{
#ifdef DEMO_MODE
		ibsta = 0;
		status = LACS;
		sleep_time = obs->sweep / 100;
		//sleep(sleep_time);
#else
		res = ibrsp(dev, &status);
		PVERB(PINFO,
                  "# ibrsp->status(%d): 0x%x wait for 0x%x, ibsta: 0x%x\n", 
			i, status, LACS, res);
		
#endif
		if(res & ERR){
			SYSLOG(LOG_ERR,"ibsta & ERR\n");
			return -1;			
		}
		if( status & LACS){
			sweep_on = 0;
		}else{
			sleep(1);
		}
		if( i > 5) {
		  SYSLOG(LOG_ERR,"Sweep time expired.\n");
		  return -1;
		}
		i++;
	}while(sweep_on);
#ifndef DEMO_MODE
	sleep(1);
#endif
	MY_IBWRT(dev,"TRA?;");
	
#ifdef DEMO_MODE
	for(i=0;i<AN_MAX_POINTS-1;i++){
		sprintf(buff+i*8,"%7.2f,", rand() / (RAND_MAX+1.0) * -100.0);
	}
	sprintf(buff+i*8,"%7.2f%c", rand() / (RAND_MAX+1.0) * -100.0,0);
#else
	ibrd(dev, buff, (AN_MAX_POINTS << 3) + 2);
#endif
	/* bytes_read = ibcnt; */
	PVERB(PINFO,"read: %d bytes\n", ibcnt);
	
	for(i=0;i<AN_MAX_POINTS;i++){
		*(s->data+i) =  strtod(np, &endp);
		np = endp + 1;
		//printf("%7.2f\n", *(s->data+i));
        }
	
	return 0;
}

#endif
